:-module(dtdtypes,[list_types/2,imprime/1,list_tupl2/2]).
:-use_module(replace).
:-op(800,xfx,--->).

%-------------------------------------------------------------------------
%
%   DTD TYPES
%
% by Jorge Coelho  jcoelho@dei.isep.ipp.pt
%    Mrio Florido amf@ncc.up.pt
%
%-------------------------------------------------------------------------

list_types(FichDTD,ListFinal1):-
	new_dtd(nome_dtd,DTD), 
	load_dtd(DTD,FichDTD),
	findall((X
,Y),dtd_property(DTD,element(X,_,Y)),List),
	process_list_types(List,DTD,ListA,Root),
	process_type_op(ListA,ListFinal),
	internal_notation(ListFinal,Root,ListFinal1).


process_list_types([(X,T)],_,[X--->[TermType] | SubTypeList],X):-!,
	process_type(T,FT),
	append([X],FT,ListTerm),
	TermType=..ListTerm,
	process_type2(T,SubTypeList).

process_list_types([(X,T)|R],DTD,[X--->[TermType] | RFinal],Root):-
	process_type(T,FT),
	append([X],FT,ListTerm),
	TermType=..ListTerm,
	process_type2(T,SubTypeList),!,
	append(SubTypeList,R2,RFinal),
	process_list_types(R,DTD,R2,Root).


process_type(empty,[nil]):-!.
process_type('#pcdata',[string]):-!.
process_type('#cdata',[string]):-!.
process_type((X,Y),[X|R]):-!,
	process_type(Y,R).
process_type(X,[X]).

process_type2(empty,[]):-!.
process_type2('#pcdata',[]):-!.
process_type2('#cdata',[]):-!.

process_type2(+((X,Y)),[+((X,Y))--->[.(((X,Y)),nil),.(((X,Y)),+((X,Y)))]|R]):-!,
	process_type2(X,R1),
	process_type2(Y,R2),	
	append(R1,R2,R).
process_type2(+(X),[+(X)--->[.(X,nil),.(X,+(X))]]):-!.

process_type2(*((X,Y)),[*((X,Y))--->[nil,.(((X,Y)),*((X,Y)))]|R]):-!,
	process_type2(X,R1),
	process_type2(Y,R2),	
	append(R1,R2,R).

process_type2(*(X),[*(X)--->[nil,.(X,*(X))]]):-!.


process_type2(?((X,Y)),R):-!,
	process_type2(X,R1),
	process_type2(Y,R2),	
	append(R1,R2,R).	

process_type2(?(_),[]):-!.


process_type2((X|Y),ListTypes):-!,
	process_type_or((X|Y),(X|Y),ListTypes).

process_type2((X,Y),R):-!,
	process_type2(X,R1),
	process_type2(Y,R2),
	append(R1,R2,R).

process_type2(_,[]).


process_type_or(X,Type,[Type--->L]):-
	process_type_or2(X,L).

process_type_or2((X|Y),[X|RType]):-!,
	process_type_or2(Y,RType).
process_type_or2((X),[X]).


imprime([]).
imprime([X|R]):-
	nl,write('---> '),write(X),imprime(R).


process_type_op([],[]).


process_type_op([X--->[Y]|R1],[X--->[Y]|R2]):-
	atom(Y),!,
	process_type_op(R1,R2).


process_type_op([X--->[T]|R1],[X--->[FinalT]|R2]):-
	T=..[X,?(Y)],!,
	Term1=..[X,Y],
	Term2=[X,nil],
	append([Term1],[Term2],FinalT),
	process_type_op(R1,R2).

process_type_op([X--->[T]|R1],[X--->[T]|R2]):-
	T=..[X,_],!,
	process_type_op(R1,R2).


process_type_op([(X|Y)--->L|R1],[(X|Y)--->L|R2]):-!,
	process_type_op(R1,R2).


process_type_op([X--->[T]|R1],[X--->L1|R2]):-
	T=..[X|R],
	process_topt2(X,R,L1),
	process_type_op(R1,R2).


process_type_op([+(X)--->L1|R1],[+(X)--->L1|R2]):-
	tupl_list((X),ListX),
	all_opt(ListX,ListOpt), 
	combina(ListOpt,[[]]),!,
	process_type_op(R1,R2).


process_type_op([+(X)--->_|R1],[+(X)--->NewList|R2]):-!,
	tupl_list((X),ListX), 
	all_opt(ListX,ListOpt),
	combina(ListOpt,ListOpt2),
	process_topt4(ListX,ListOpt2,NewList),
	process_type_op(R1,R2).


process_type_op([*(X)--->L1|R1],[*(X)--->L1|R2]):-
	tupl_list((X),ListX),
	all_opt(ListX,ListOpt), 
	combina(ListOpt,[[]]),!,
	process_type_op(R1,R2).


process_type_op([*(X)--->_|R1],[*(X)--->NewList2|R2]):-!,
	tupl_list((X),ListX), 
	all_opt(ListX,ListOpt),
	combina(ListOpt,ListOpt2),
	process_topt5(ListX,ListOpt2,NewList),
        append([nil],NewList,NewList2),
	process_type_op(R1,R2).

process_topt2(X,[Y|R],[Term]):-
	all_opt([Y|R],ListOpt),
	combina(ListOpt,[]),!,
	append([X],[Y|R],NewList),
	Term=..NewList.

process_topt2(X,[Y|R],NewList):-
	all_opt([Y|R],ListOpt),
	combina(ListOpt,ListOpt2),
	process_topt3(X,[Y|R],ListOpt2,NewList).


process_topt3(_,_,[],[]).

process_topt3(X,[Y|Z],[E|R1],[Term|R2]):-
	new_opt_list([Y|Z],E,L),
	append([X],L,TermList),
	Term=..TermList,
	process_topt3(X,[Y|Z],R1,R2).



process_topt4(_,[],[]).

process_topt4(ListX,[E|R2],[[Tupl|nil],[Tupl|Term]|RFinal]):-
	new_opt_list(ListX,E,L),
	list_tupl(L,Tupl),
	append([+],ListX,TermList),
	Term=..TermList,
	process_topt4(ListX,R2,RFinal).

process_topt5(_,[],[]).

process_topt5(ListX,[E|R2],[ [Tupl|Term] |RFinal]):-
	new_opt_list(ListX,E,L),
	list_tupl(L,Tupl),
	list_tupl2(ListX,TX),
	append([*],[TX],TermList),
	Term=..TermList,
	process_topt5(ListX,R2,RFinal).


del(X,[X|R],R).
del(X,[Y|R1],[Y|R2]):-
	del(X,R1,R2).


my_insert(X,L,L2):-
	del(X,L2,L).


tupl_list((X,Y),[X|R]):-!,
	tupl_list(Y,R).

tupl_list((X),[X]).


list_tupl([X],(X)).
list_tupl([?(X)|R],((X,R2))):-!,
	list_tupl(R,R2).
list_tupl([X|R],((X,R2))):-
	list_tupl(R,R2).


list_tupl2([X],(X)).
list_tupl2([X|R],((X,R2))):-
	list_tupl2(R,R2).


new_opt_list([],_,[]).

new_opt_list([?(Y)|Z],E,R):-
	member(Y,E),!,
	new_opt_list(Z,E,R).

new_opt_list([?(Y)|Z],E,RFinal):-
	not(member(Y,E)),!,
	tupl_list(Y,LY),
	new_opt_list(Z,E,R),
	new_type_list(LY,R,RFinal).

new_opt_list([Y|Z],E,[Y|R]):-
	new_opt_list(Z,E,R).


all_opt([?(X)|R1],[X|R2]):-!,
	all_opt(R1,R2).
all_opt([_|R1],R2):-!,
	all_opt(R1,R2).
all_opt([?(X)],[X]):-!.
all_opt([_],[]).	
all_opt([],[]).


combina([],[]).
combina(L1,L4):-
	findall(X,select(X,L1,_),List1),
	convert(List1,List11),
	findall(Y,select(_,L1,Y),List2),
	append(List11,List2,L2),
	append([L1],L2,L3),
	append([[]],L3,L31), % elemento vazio  necessrio para a criao de termo sem opcionais.
	setof(X,member(X,L31),L4).


convert([],[]).
convert([X|Y],[[X]|RY]):-
	convert(Y,RY).
	

internal_notation(L1,Root,L5):-
	tupls_list(L1,L2),
	type_notation(L2,L3),
	term_to_atom(Root,Root1),
	root_type(L3,Root1,L4),
	add_attribute_type(L4,L5).


root_type([type_(Root,X)--->Y|R],Root,[type_(root,X)--->Y|R]):-!.

root_type([X--->Y|R1],Root,[X--->Y|R2]):-
	root_type(R1,Root,R2).


tupls_list([],[]).

tupls_list([+X--->L1|R1],[+X--->L2|R2]):-!,
	tupls_list2(L1,L2),
	tupls_list(R1,R2).

tupls_list([*(X)--->L1|R1],[*(X)--->L2|R2]):-!,
	tupls_list2(L1,L2),
	tupls_list(R1,R2).

tupls_list([X--->L|R1],[X--->L|R2]):-
	tupls_list(R1,R2).


tupls_list2([],[]).

tupls_list2([nil|R1],[nil|R2]):-
	tupls_list2(R1,R2).

tupls_list2([[A|B]|R1],[C|R2]):-
	tupl_list(A,A1),
	new_type_list(A1,B,C),
	tupls_list2(R1,R2).

new_type_list([],B,B).

new_type_list([A|R1],B,[A|R2]):-
	new_type_list(R1,B,R2).



type_notation(L1,L3):-
	all_atom(L1,L2),
	internal_types(L2,L3).

all_atom(L1,L3):-
	create_subs(L1,L2,S),
	sub_rules(L2,S,L3).


create_subs([],[],[]).

create_subs([X--->L|R1],[Y--->L|R2],[(X,Y)|R3]):-
	term_to_atom(X,Y),
	create_subs(R1,R2,R3).


sub_rules([],_,[]).

sub_rules([X--->Y|R1],S,[X--->Z|R2]):-
	replace(Y,S,Z,[],_),
	sub_rules(R1,S,R2).

internal_types(L1,L3):-
	internal_types2(L1,L2,S),
	apply_subs_rules(L2,S,L3).

internal_types2([],[],[(nil,[])]).

internal_types2([X--->Y|R1],[type_(X,0)--->Y|R2],[(X,type_(X,0))|R3]):-
	internal_types2(R1,R2,R3).

apply_subs_rules([],_,[]).
apply_subs_rules([X--->Y1|R1],S,[X--->Y2|R2]):-
	replace(Y1,S,Y2,[],_),
	apply_subs_rules(R1,S,R2).

add_attribute_type([],[]).

add_attribute_type([X--->Y1|R1],[X--->Y2|R2]):-
	add_attribute_type2(Y1,Y2),
	add_attribute_type(R1,R2).

add_attribute_type2([],[]).

add_attribute_type2([A|R1],[B|R2]):-
	A=..[F|RA],
	F\='.',
	F\=[],
	F\='type_',!,
	B=..[F,any|RA],
	add_attribute_type2(R1,R2).

add_attribute_type2([A|R1],[A|R2]):-
	add_attribute_type2(R1,R2).


