:-module(infertypes,[infer_types/4,tuple_to_list/2,built_in_temp/1,my_subset/3]).
:-use_module(replace).

%-------------------------------------------------------------------------
%
%  TYPE INFERENCE
%
% by Jorge Coelho  jcoelho@dei.isep.ipp.pt
%    Mrio Florido amf@ncc.up.pt
%
%
%-------------------------------------------------------------------------


:-op(800,xfx,--->).

built_in_temp([type_(==,1)--->[expr],type_(==,2)--->[expr],type_(>,1)--->[expr],type_(>,2)--->[expr],type_(<,1)--->[expr],type_(<,2)--->[expr],type_(=<,1)--->[expr],type_(=<,2)--->[expr],type_(>=,1)--->[expr],type_(>=,2)--->[expr],type_(\==,1)--->[expr],type_(\==,2)--->[expr],type_(+,1)--->[expr],type_(+,2)--->[expr],type_(-,1)--->[expr],type_(-,1)--->[expr],type_(*,1)--->[expr],type_(*,2)--->[expr],type_(/,1)--->[expr],type_(/,2)--->[expr],type_(is,1)--->[number],type_(is,2)--->[expr]]).


empty(none,_,_).


empty(X,E,_):-member(X,E).


empty(type_(P,X),E,TRules):-
	member(type_(P,X)--->L,TRules),
	union([type_(P,X)],E,NE),
	empty_tr(L,NE,TRules).


empty(F,E,TRules):-
	F=..[_|RLX],
	empty_tt(RLX,E,TRules).


empty_tt([],_,_):-fail.

empty_tt([A|RA],E,TRules):-
	not(empty(A,E,TRules)),
	empty_tt(RA,E,TRules).


empty_tt([A|_],E,TRules):-
	empty(A,E,TRules).


empty_tr([],_,_).


empty_tr([Y|RL],E,TRules):-
	empty(Y,E,TRules),
	empty_tr(RL,E,TRules).


base_type(int).
base_type(string).
base_type(const).
base_type(number).
base_type(expr).


type_of(const,X):-atom(X).
type_of(string,X):-atom(X).
type_of(number,X):-number(X).
type_of(expr,X):-number(X).
type_of(expr,X):-expr(X).
type_of(int,X):-integer(X).




expr(var_(_)).
% necessrio para a interseco.
expr(any).
expr(X):-number(X).
expr(X+Y):-expr(X),expr(Y).
expr(X-Y):-expr(X),expr(Y).
expr(X*Y):-expr(X),expr(Y).
expr(X/Y):-expr(X),expr(Y).



intersection(T1,T2,TRules,TF,TFRules):-
	intersectionv(T1,T2,TRules,[],TF,TFRules,1,_).


intersectionv(any,T,TRules,_,T,TRules,N,N):-!.
intersectionv(T,any,TRules,_,T,TRules,N,N):-!.


intersectionv(T,T,TRules,_,T,TRules,N,N):-!.


intersectionv(T1,T2,TRules,I,Type,TRules,N,N):-
	member([T1,T2,Type],I),!.

intersectionv(T1,T2,TRules,I,Type,TRules,N,N):-
	member([T2,T1,Type],I),!.

intersectionv(T1,T2,TRules,_,T2,TRules,N,N):-
	base_type(T1),
	type_of(T1,T2),!.
intersectionv(T1,T2,TRules,_,T1,TRules,N,N):-
	base_type(T2),
	type_of(T2,T1),!.

intersectionv(type_(P1,T1),type_(P2,T2),TRules,I,type_(0,N1),TRules3,N1,N3):-
	type_defenition(type_(P1,T1),TRules,TDef1),
	type_defenition(type_(P2,T2),TRules,TDef2),!,
	union(I,[[type_(P1,T1),type_(P2,T2),type_(0,N1)]],NI),
	N2 is N1+1,
	cpi(TDef1,TDef2,TRules,NI,TDef3,TRules2,N2,N3),
	clean_nones(N1,TDef3,TDef4),
	union(TRules2,[type_(0,N1)--->TDef4],TRules3).

intersectionv(type_(P1,T1),T2,TRules,I,type_(0,N1),TRules3,N1,N3):-
	T2=..[Temp|_],
	Temp\=type_,
	type_defenition(type_(P1,T1),TRules,TDef1),
	type_defenition(T2,TRules,TDef2),
	union(I,[[type_(P1,T1),T2,type_(0,N1)]],NI),
	N2 is N1+1,
	cpi(TDef1,TDef2,TRules,NI,TDef3,TRules2,N2,N3),
	clean_nones(N1,TDef3,TDef4),
	TDef4\=[],!,
	union(TRules2,[type_(0,N1)--->TDef4],TRules3).

intersectionv(T1,type_(P2,T2),TRules,I,type_(0,N1),TRules3,N1,N3):-
	T1=..[Temp|_],
	Temp\=type_,
	type_defenition(type_(P2,T2),TRules,TDef1),
	type_defenition(T1,TRules,TDef2),
	union(I,[[type_(P2,T2),T1,type_(0,N1)]],NI),
	N2 is N1+1,
	cpi(TDef1,TDef2,TRules,NI,TDef3,TRules2,N2,N3),
	clean_nones(N1,TDef3,TDef4),
	TDef4\=[],!,
	union(TRules2,[type_(0,N1)--->TDef4],TRules3).

intersectionv(F1,F2,TRules,I,T,TFRules,N1,N2):-
	F1=..[F|L1],
	F2=..[F|L2],
	F\=type_,!,
	list_intersection(L1,L2,TRules,I,LF,TFRules,N1,N2),
	T=..[F|LF].

intersectionv(_,_,TRules,_,none,TRules,N,N).




list_intersection([],_,TRules,_,[],TRules,N,N).

list_intersection([],[_|_],TRules,_,[none],TRules,N,N).

list_intersection([_|_],[],TRules,_,[none],TRules,N,N):-!.

list_intersection([X|RX],[Y|RY],TRules,I,[XY|RXY],TRules3,N1,N3):-
	intersectionv(X,Y,TRules,I,XY,TRules2,N1,N2),
	list_intersection(RX,RY,TRules2,I,RXY,TRules3,N2,N3).


type_defenition(T,TRules,L):-
	member(T--->L,TRules),!.

type_defenition(T,_,[T]).



cpi([],_,TRules,_,[],TRules,N,N).

cpi([X|R],L,TRules,I,TDef3,TRules3,N1,N3):-
	cpi2(X,L,TRules,I,TDef,TRules2,N1,N2),
	cpi(R,L,TRules2,I,TDef2,TRules3,N2,N3),
	union(TDef,TDef2,TDef3).




cpi2(_,[],TRules,_,[],TRules,N,N).

cpi2(X,[Y|R],TRules,I,[TF|R2],TRules2,N1,N3):-
	intersectionv(X,Y,TRules,I,TF,TRulesF,N1,N2),
	cpi2(X,R,TRulesF,I,R2,TRules2,N2,N3).



clean_nones(_,[],[]).

clean_nones(N,[X|RX],RY):-
	empty(X,[],[]),!,
	clean_nones(N,RX,RY).

clean_nones(N,[type_(_,N)|RX],RY):-!,
	clean_nones(N,RX,RY).

clean_nones(N,[X|RX],[X|RY]):-
	clean_nones(N,RX,RY).

list_print([]). 

list_print([X|R]):-
	nl,write(X),
	list_print(R).


unify(X,X,TRules,Subst,X,TRules,Subst,N,N):-!.

unify(any,X,TRules,Subst,X,TRules,Subst,N,N):-!.

unify(X,any,TRules,Subst,X,TRules,Subst,N,N):-!.

unify(var_(X),var_(Y),TRules,Subst,var_(X),TRules2,Subst3,N1,N2):-
	X<Y,!,
	member((var_(X),SubX),Subst),
	unify(SubX,var_(Y),TRules,Subst,Z,TRules2,Subst2,N1,N2),
	subst((var_(X),SubX),(var_(X),Z),Subst2,Subst3).

unify(var_(X),var_(Y),TRules,Subst,var_(Y),TRules2,Subst3,N1,N2):-
	Y<X,!,
	member((var_(Y),SubY),Subst),
	unify(SubY,var_(X),TRules,Subst,Z,TRules2,Subst2,N1,N2),
	subst((var_(Y),SubY),(var_(Y),Z),Subst2,Subst3).

unify(var_(X),Y,TRules,Subst,var_(X),TRules2,Subst3,N1,N2):-
	Y=..[C|R],
	C\=var_,
	not(member(var_(X),[C|R])),!,
	member((var_(X),SubX),Subst),
	unify(SubX,Y,TRules,Subst,Z,TRules2,Subst2,N1,N2),
	subst((var_(X),SubX),(var_(X),Z),Subst2,Subst3).

unify(var_(X),Y,TRules,Subst,none,TRules,Subst,N,N):-
	Y=..[C|R],
	C\=var_,!,
	member(var_(X),[C|R]).

unify(X,var_(Y),TRules,Subst,var_(Y),TRules2,Subst3,N1,N2):-
	X=..[C|R],
	C\=var_,
	not(member(var_(Y),[C|R])),!,
	member((var_(Y),SubY),Subst),
	unify(SubY,X,TRules,Subst,Z,TRules2,Subst2,N1,N2),
	subst((var_(Y),SubY),(var_(Y),Z),Subst2,Subst3).

unify(X,var_(Y),TRules,Subst,none,TRules,Subst,N,N):-
	X=..[C|R],
	C\=var_,!,
	member(var_(Y),[C|R]).

unify(X,Y,TRules,Subst,Z,TRules2,Subst,N1,N2):-
	pure_type_term(X),
	pure_type_term(Y),!,
	intersectionv(X,Y,TRules,[],Z,TRules2,N1,N2).


unify(type_(P,X),Y,TRules,Subst,ZF,TRules5,Subst2,N1,N5):-
	member(type_(P,X)--->_,TRules),
	Y=..[F|R],!,
	universal(R,R1),
	T=..[F|R1],
	union(TRules,[type_(0,N1)--->[T]],TRules2),
	N2 is N1+1,
	intersectionv(type_(0,N1),type_(P,X),TRules2,[],Z,TRules3,N2,N3),
	member(Z--->L2,TRules3),
	new_types(L2,TRules3,L3,N3,N4),
	new_struct(F,L3,NS),
	union(TRules3,L3,TRules4),
	unify(Y,NS,TRules4,Subst,ZF,TRules5,Subst2,N4,N5).

unify(X,type_(P2,Y),TRules,Subst,ZF,TRules5,Subst2,N1,N5):-
	member(type_(P2,Y)--->_,TRules),
	X=..[F|R],!,
	universal(R,R1),
	T=..[F|R1],
	union(TRules,[type_(0,N1)--->[T]],TRules2),
	N2 is N1+1,
	intersectionv(type_(0,N1),type_(P2,Y),TRules2,[],Z,TRules3,N2,N3),
	member(Z--->L2,TRules3),
	new_types(L2,TRules3,L3,N3,N4),
	new_struct(F,L3,NS),
	union(TRules3,L3,TRules4),
	unify(X,NS,TRules4,Subst,ZF,TRules5,Subst2,N4,N5).

unify(X,Y,TRules,Subst,Z,TRules2,Subst2,N1,N2):-
	X=..[F|R1],
	Y=..[F|R2],
	F\=type_,
	F\=var_,!,
	m_unify(R1,R2,TRules,Subst,R3,TRules2,Subst2,N1,N2),
	Z=..[F|R3].



unify(_,_,TRules,Subst,none,TRules,Subst,N,N).



subst(_,_,[],[]).

subst(X,Y,[X|R1],[Y|R2]):-
	subst(X,Y,R1,R2).

subst(X,Y,[Z|R1],[Z|R2]):-
	X\=Z,
	subst(X,Y,R1,R2).


pure_type_term(X):-
	X=..L,
	no_vars(L).



no_vars([]).

no_vars([X|R]):-
	X=..[F|R2],
	F\=var_,
	no_vars(R2),
	no_vars(R).


new_struct(F,L1,FE):-
	new_struct2(L1,L2),
	FE =..[F|L2].


new_struct2([],[]).

new_struct2([T--->_|RT1],[T|RT2]):-
	new_struct2(RT1,RT2).


universal([],[]).

universal([_|R1],[any|R2]):-
	universal(R1,R2).


new_types([type_(X,Y)],TR,LT,N1,N2):-
	member(type_(X,Y)--->Z,TR),
	new_types(Z,TR,LT,N1,N2).

new_types(LF,_,LT,N1,N2):-
	extract_args(LF,LF2),
	union_args(LF2,LF3),
	generate_types(LF3,LT,N1,N2).



generate_types([],[],N,N).

generate_types([LT|RLT],[type_(0,N1)--->LT2|RNT],N1,N3):-
	N2 is N1 + 1,
	list_to_set(LT,LT2),
	generate_types(RLT,RNT,N2,N3).


heads([],[],[]).
heads([[]|_],[],[]).
heads([[H1|RL]|R1],[H1|R2],[RL|R3]):-
	heads(R1,R2,R3).



empty_lists([]).

empty_lists([[]|R]):-
	empty_lists(R).


union_args(L,[]):-
	empty_lists(L),!.

union_args(L1,L4):-
	heads(L1,L2,RL1),
	union_args(RL1,L3),
	append([L2],L3,L4).


extract_args([],[]).

extract_args([X|R1],[Y|R2]):-
	X=..[_|Y],
	extract_args(R1,R2).


m_unify([],[],TRules,Subst,[],TRules,Subst,N,N).

m_unify([X|RX],[Y|RY],TRules,Subst,[XY|RXY],TRules3,Subst3,N1,N3):-
	unify(X,Y,TRules,Subst,XY,TRules2,Subst2,N1,N2),
	m_unify(RX,RY,TRules2,Subst2,RXY,TRules3,Subst3,N2,N3).




goaltype(G,T,V,T2,Subst2,N1,N2):-
	subst_univ(V,Subst),
	goaltype2(G,T,Subst,T2,Subst2,N1,N2).




goaltype2(_,[],TRules,Subst,TRules,Subst,N,N).

goaltype2(X,[G1=G2|RG],TRules,Subst,TRules3,Subst3,N1,N3):-
	unify(G1,G2,TRules,Subst,G3,TRules2,Subst2,N1,N2),
	substitution(G3,Subst2,S,[]),
        not(empty(S,[],TRules2)),!,
	goaltype2(X,RG,TRules2,Subst2,TRules3,Subst3,N2,N3).

goaltype2(X,[G1|RG],TRules,Subst,TRules3,Subst3,N1,N3):-
        G1\=(_=_),
	G1=..[F|R],
	type_def(F,R,GR,1),
	G2=..[F|GR],
	unify(G1,G2,TRules,Subst,G3,TRules2,Subst2,N1,N2),
	substitution(G3,Subst2,S,[]),
        not(empty(S,[],TRules2)),!,
	goaltype2(X,RG,TRules2,Subst2,TRules3,Subst3,N2,N3).




subst_univ([],[]).
subst_univ([X|R],[(X,any)|R2]):-
	subst_univ(R,R2).


type_def(_,[],[],_).

type_def(F,[_|R],[type_(F,N)|R2],N):-
	N1 is N + 1,
	type_def(F,R,R2,N1).



substitution(G,_,G,_):-
	pure_type_term(G),!.

substitution(G,Subst,G,E):-
	replace(G,Subst,_,E,[]).

substitution(G,Subst,G3,E1):-
	replace(G,Subst,G2,E1,E2),
	substitution(G2,Subst,G3,E2).


programtype(P,MPred,T,NT3,N1,N3):-
	programtype2(P,MPred,T,T2,FT,N1,N2),
	remove_old_types(FT,T2,T3),
	union(T3,FT,NT1),
	tuple_distribution(NT1,NT2,N2,N3),
	simplify(NT2,NT3).

simplify([],[]).

simplify([X--->Y|R1],[X--->Y1|R2]):-
	member(X,Y),!,
	delete(Y,X,Y1),
	simplify(R1,R2).

simplify([X--->Y|R1],[X--->Y|R2]):-
	simplify(R1,R2).


programtype2([],_,T,T,[],N,N).

programtype2([X|R],MPred,T1,T3,FL,N1,N3):-
	programtype3(X,MPred,T1,T2,L,N1,N2),
	final_types(X,L,L2),
	programtype2(R,MPred,T2,T3,RL,N2,N3),
	union(L2,RL,FL).


programtype3([],_,T,T,[],N,N).


programtype3([X|R],MPred,T1,T2,L,N1,N2):-
	no_args(X),!,
	programtype3(R,MPred,T1,T2,L,N1,N2).

programtype3([[X|RX]|R],MPred,T1,T3,[L2|RL],N1,N3):-
	vars([X|RX],V),
	X=..[_|L],
	programtype4(X,RX,MPred,V,T1,T2,L,L2,N1,N2),
	programtype3(R,MPred,T2,T3,RL,N2,N3).


programtype4(X,A,F,V,T1,T2,L1,L2,N1,N2):-
        X=..[F|Args],
        subs_head_vars(F,Args,NSub,1),
	subst_univ(V,Subst),
        no_rep(Subst,SubstR),
	swaps(NSub,SubstR,SubstX),
	goaltype2(X,A,T1,SubstX,T2,Subst2,N1,N2),!,
	substitution(L1,Subst2,L2,[]).

programtype4(X,A,F1,V,T1,T2,L1,L2,N1,N2):-
        X=..[F|_],
	F \= F1,
	subst_univ(V,Subst),
        no_rep(Subst,SubstR),
	goaltype2(X,A,T1,SubstR,T2,Subst2,N1,N2),!,
	substitution(L1,Subst2,L2,[]).


subs_head_vars(_,[],[],_).

subs_head_vars(F,[X|RX],[(X,type_(F,N))|RY],N):-
	N1 is N + 1,
        subs_head_vars(F,RX,RY,N1).

no_rep([],[]).

no_rep([X|RX],R2):-
        no_rep(RX,R1),
	union([X],R1,R2).

swaps([],L,L).

swaps([(X,Y)|RX],L,L2):-
	swap((X,_),(X,Y),L,L1),
        swaps(RX,L1,L2).

swap(_,_,[],[]).

swap(X,Y,[X|R],[Y|R]):-!.

swap(X,Y,[Z|R1],[Z|R2]):-
    swap(X,Y,R1,R2).



remove_old_types(_,[],[]).

remove_old_types(FT,[X--->_|R1],[X--->Z|R2]):-
	member(X--->Z,FT),!,
	remove_old_types(FT,R1,R2).

remove_old_types(FT,[X--->Y|R1],[X--->Y|R2]):-
	remove_old_types(FT,R1,R2).



final_types([[X|_]|_],L,L2):-
	X=..[F|_],
	final_types2(F,L,L2,1).


final_types2(_,L,[],_):-
	empty_lists(L),!.

final_types2(F,L,[type_(F,N)--->L2|RL],N):-
	N1 is N+1,
	heads(L,L2,L3),
	final_types2(F,L3,RL,N1).


add_nones([],[]).

add_nones([_|R1],[none|R2]):-
	add_nones(R1,R2).


no_args([X|_]):-
	atom(X).


tuple_distribution([],[],N,N).

tuple_distribution([type_(X,Y)--->L|R1],[type_(X,Y)--->L8|R4],N1,N4):-
	groups2(L,L,L2),
	duplicates(L2,L3,L4),
	clean_no_dups(L4,L,L42),
	heads(L3,L5,_),
	nt(L3,L6,N1,N2),
	flatten(L6,L63),
	tuple_distribution(L63,L64,N2,N3),
	final_types3(L5,L6,L7),
	union(L42,L7,L8),
	tuple_distribution(R1,R2,N3,N4),
	union(L64,R2,R3),
	flatten(R3,R4).



clean_no_dups([],_,[]).

clean_no_dups([[X]|RX],L,[X|RY]):-
	member(X,L),!,
	clean_no_dups(RX,L,RY).

clean_no_dups([X|RX],L,[X|RY]):-
	clean_no_dups(RX,L,RY).


final_types3([],[],[]).

final_types3([X|RX],[Y|RY],[NT|RNT]):-
	create_subs(Y,Z),
	subst_i(X,Z,NT),
	final_types3(RX,RY,RNT).



subst_i(F,TL,F2):-
	F=..[A|AS],
	subst_i2(AS,TL,TL,NAS),
	F2=..[A|NAS].


subst_i2([],_,_,[]).

subst_i2([X|RX],LT,LT3,[NX|NRX]):-
	subst_i3(X,LT,LT2,NX),
	subst_i2(RX,LT2,LT3,NRX).



subst_i3(F,[],[],F).

subst_i3(X,[(X,Y)|RXY],RXY,Y).

subst_i3(X,[(Z,T)|RXY],[(Z,T)|RXY2],Y):-
	X\=Z,
	subst_i3(X,RXY,RXY2,Y).


create_subs([],[]).

create_subs([type_(X,Y)--->L|RL],S3):-
	create_subs2(L,type_(X,Y),S1),
	create_subs(RL,S2),
	union(S1,S2,S3).
	
create_subs2([],_,[]).

create_subs2([X|RX],Y,[(X,Y)|RS]):-
	create_subs2(RX,Y,RS).



duplicates([],[],[]).

duplicates([X|R1],[X|R2],R3):-
	length(X,N),
	N>1,!,
	duplicates(R1,R2,R3).

duplicates([X|R1],R2,[X|R3]):-
	is_list(X),
	duplicates(R1,R2,R3).




groups2([],_,[]).

groups2([X|R],L,LF2):-
	one_element(X),!,
	groups2(R,L,RLF),
	union([[X]],RLF,LF2).

groups2([type_(X,Y)|R],L,LF2):-!,
	groups2(R,L,LF),
	union([[type_(X,Y)]],LF,LF2).

groups2([X|R],L,LF2):-
	X=..[F|RF],
	not(is_list(X)),!,
	length(RF,N),
	groups3(F,N,L,LF),
	groups2(R,L,RLF),
	union([LF],RLF,LF2).


groups2([X|R],L,LF2):-
	X=..[_|_],
	is_list(X),
	groups2(R,L,RLF),
	union([X],RLF,LF2).


groups3(_,_,[],[]).

groups3(F,N,[X|RX],[X|RY]):-
	X=..[F|R],
	length(R,N),!,
	groups3(F,N,RX,RY).

groups3(F,N,[_|RX],RY):-
	groups3(F,N,RX,RY).


one_element(X):-
	X=..[X],
	not(is_list(X)),!.

one_element([X]):-
	one_element(X).


nt([],[],N,N).

nt([X|R],[NT|RNT],N1,N3):-
	new_types(X,_,NT,N1,N2),
	nt(R,RNT,N2,N3).


head([],[]).

head([X|_],X).



heads2([],[]).

heads2([X|RX],[Y|RY]):-
	head(X,Y),
	heads2(RX,RY).


tail([],[]).

tail([_|R],R).


tails([],[]).

tails([X|RX],[Y|RY]):-
	tail(X,Y),
	tails(RX,RY).


expand([],_,[]).

expand([H|L],_,[[H|L]]):-
	H=..[F|_],
	F\=type_.

expand([H|L],T,L3):-
	member(H--->L2,T),
	sequences(L2,L,L3).


sequences([],_,[]).

sequences([X|RX],L,[Y|RY]):-
	union([X],L,Y),
	sequences(RX,L,RY).


expands([],_,[]).

expands([X|RX],T,R):-
	expand(X,T,Y),
	expands(RX,T,RY),
	union(Y,RY,R).


expands2(X,T,E,L2):-
	expands(X,T,L),
	expands3(L,T,E,L2).

expands3([],_,_,[]).

expands3([X|RX],T,E,L):-
	member(X,E),!,
	expands3(RX,T,E,L).

expands3([X|RX],T,E,L2):-
	expands([X],T,[X]),!,
	expands3(RX,T,E,L1),
	union([X],L1,L2).

expands3([X|RX],T,E,L2):-
	expands([X],T,Y),
	union([X],E,E2),
	expands2(Y,T,E2,Y2),
	expands3(RX,T,E2,L1),
	union(Y2,L1,L2).



selects(any,TS,T):-
	member_heads(any,TS,T),!.

selects(X,TS,T3):-
	base_type(X),!,
	member_heads(X,TS,T1),
	member_heads(any,TS,T2),
	union(T1,T2,T3).

selects(X,TS,T5):-
	atom(X),!,
	member_heads(X,TS,T1),
	member_heads(any,TS,T2),
	member_type(X,TS,T3),
	union(T1,T2,T4),
	union(T3,T4,T5).


selects(X,TS,T5):-
	integer(X),!,
	member_heads(X,TS,T1),
	member_heads(any,TS,T2),
	member_type(X,TS,T3),
	union(T1,T2,T4),
	union(T3,T4,T5).

selects(X,TS,T3):-
	X=..[_|R],
	length(R,N),
	N>0,
	member_heads(any,TS,T1),
	member_heads_func(X,TS,T2),
	union(T1,T2,T3).


member_heads_func(_,[],[]).

member_heads_func(H1,[X|R1],[X|R2]):-
	H1=..[F|RH1],
	head(X,H2),
	H2=..[F|RH2],
	length(RH1,N),
	length(RH2,N),!,
	member_heads_func(H1,R1,R2).

member_heads_func(H,[_|R1],R2):-
	member_heads_func(H,R1,R2).



member_type(X,TS,T3):-
	type_of(B,X),
	member_heads(B,TS,T3).


member_heads(_,[],[]).

member_heads(H,[X|R1],[X|R2]):-
	head(X,H),
	member_heads(H,R1,R2).

member_heads(H,[X|R1],R2):-
	head(X,G),
	G\=H,
	member_heads(H,R1,R2).


member_heads2(_,[],[]).

member_heads2(H,[X|R1],[X|R2]):-
	head(X,G),
	G=..[type_|_],
	member_heads2(H,R1,R2).

member_heads2(H,[X|R1],R2):-
	head(X,G),
	G\=[type_|_],
	member_heads2(H,R1,R2).




my_open(X,TS,Y):-
	atom(X),
	tail(TS,Y).

my_open(X,TS,Y):-
	integer(X),
	tail(TS,Y).

my_open(X,TS,Y):-
	base_type(X),
	tail(TS,Y).

my_open(X,TS,Y):-
	X=..[F|R],
	F\=var_,
	F\=type_,
	head(TS,any),
	length(R,N),
	N>0,
	universal(R,U),
	tail(TS,Z),
	union(U,Z,Y).

my_open(X,TS,Y):-
	X=..[F|R],
	F\=var_,
	F\=type_,
	head(TS,H),
	H=..[F|R2],
	length(R,N),
	length(R2,N),
	N>0,
	tail(TS,Z),
	union(R2,Z,Y).
	

my_opens(_,[],[]).

my_opens(X,[T|RT],Z):-
	my_open(X,T,Y),
	my_opens(X,RT,RY),
	union([Y],RY,Z).


my_subset(none,_,_).

my_subset(_,none,_):-fail.

my_subset(T1,T2,T):-
	subsetv([T1],[[T2]],T,[]).


subsetv(_,[],_,_):-!,fail.

subsetv([],_,_,_):-!.

subsetv(L1,L2,T,C):-
	head(L1,H),
	member([H,TR],C),
	heads2(L2,H2),
	union(TR,_,H2),!,
	tail(L1,L3),
	tails(L2,L4),
	subsetv(L3,L4,T,C).

subsetv(L1,L2,T,C):-
	head(L1,H),
	H=..[type_|_],!,
	expand(L1,T,L3),
	heads2(L2,HS),
	union([[H,HS]],C,C2),
	m_subsetv(L3,T,L2,C2).

subsetv(L1,L2,T,C):-
	head(L1,H),
	test(H),!,
	my_open(H,L1,L3),
	expands2(L2,T,[],L4),
	selects(H,L4,L5),
	my_opens(H,L5,L6),
	subsetv(L3,L6,T,C).


test(X):-
	base_type(X).

test(X):-
	atom(X).

test(X):-
	integer(X).

test(any).

test(X):-
	X=..[_|R],
	length(R,N),
	N>0.

m_subsetv([],_,_,_).

m_subsetv([X|RX],T,L2,C):-
	subsetv(X,L2,T,C),!,
	m_subsetv(RX,T,L2,C).


supertypes(P,MPred,T0,T1,N1,N2):-
	not(programtype(P,MPred,T0,T1,N1,N2)),!,
	nl, write('ERROR: Incompatible types.'),nl,!,fail.

supertypes(P,MPred,T0,T1,N1,N2):-
	programtype(P,MPred,T0,T1,N1,N2),
	reached_limit(T1),!,
	nl, write("Warning: I'm unable to do type inference for this program.").


supertypes(P,MPred,T0,T0,N1,N2):-
	programtype(P,MPred,T0,T1,N1,N2),
	unic_types(T0,T00,ST0,sub1,1,NT),
	unic_types(T1,T10,ST1,sub2,NT,_),
	apply_subs(T00,ST0,T01),
	apply_subs(T10,ST1,T11),
	union(T01,T11,TS),
	my_subsett(T01,TS),
	my_subsett(T11,TS),!.

supertypes(P,MPred,T0,T2,N1,N3):-
	programtype(P,MPred,T0,T1,N1,N2),
	supertypes(P,MPred,T1,T2,N2,N3).




my_subsett([],_).

my_subsett([type_(X,Z1)--->_|R],T):-
	member(type_(Y,Z2)--->_,T),
	head_ptype(X,Y),!,
	my_subset(type_(X,Z1),type_(Y,Z2),T),
	my_subsett(R,T).

my_subsett([_|R1],R2):-
	my_subsett(R1,R2).

% verdade se X representa o mesmo tipo que Y, ex.: psub11 e psub21 representam o mesmo tipo
head_ptype(X,Y):-
	string_to_list(X,XL),
	string_to_list(Y,YL),
	string_to_list(sub,S),
	not(append(S,_,XL)),
	not(append(S,_,YL)),
	append(A,B,XL),
	append(A,C,YL),
	A\=[],
	B\=C,
	B=[_,_,_,_|RB],
	C=[_,_,_,_|RC],
	same_list(RB,RC).



unic_types([],[],[],_,N,N).

unic_types([type_(X,Y)--->TL|RT1],[type_(X1,Y)--->TL|RT2],[(type_(X,Y),type_(X1,Y))|RS],S,N,N1):-
	built_in(L),
	member(type_(X,Y)--->TL,L),!,
	term_to_atom(X,X1),           % para no dar problemas com *(a,b), etc... em head_ptype.
	unic_types(RT1,RT2,RS,S,N,N1).

unic_types([type_(0,Y)--->L|RT1],[type_(S,N)--->L|RT2],[(type_(0,Y),type_(S,N))|RS],S,N,N2):-
	N1 is N + 1,!,
	unic_types(RT1,RT2,RS,S,N1,N2).

unic_types([type_(X,Y)--->L|RT1],[type_(XS,N)--->L|RT2],[(type_(X,Y),type_(XS,N))|RS],S,N,N2):-
	X\=0,!,
	string_to_list(S,SL),
	string_to_list(X,XL),
	string_to_list(Y,NL),
	append(XL,SL,NSL),
	append(NSL,NL,NSL2),
	string_to_list(XS,NSL2),
	N1 is N + 1,
	unic_types(RT1,RT2,RS,S,N1,N2).



apply_subs([],_,[]).

apply_subs([X--->L1|RX],S,[X--->L2|RY]):-
	replace(L1,S,L2,[],_),
	apply_subs(RX,S,RY).



same_list([X],[X]).

same_list([X|RX],[X|RY]):-
	same_list(RX,RY).


read_source_file(File1,List):-
	open(File1,read,Fd),
	read(Fd,First),
	read_data(First,Fd,List),
	close(Fd).


read_data(end_of_file,_,[]):-!.

read_data((A:-B),Fd,[[A|C]|RData]):-
	tuple_to_list(B,C),
	read(Fd,Next),!,
	read_data(Next,Fd,RData).

read_data((:-_),Fd,RData):-
	read(Fd,Next),!,
	read_data(Next,Fd,RData).

read_data((A),Fd,[[A]|RData]):-
	read(Fd,Next),
	read_data(Next,Fd,RData).


tuple_to_list((A,B),[A|LB]):-!,
	tuple_to_list(B,LB).

tuple_to_list((A),[A]).




convert_program(P1,[C2|RP]):-
	convert_program2(P1,[C1|RP]),
	explicit_unification(C1,C2).
	

convert_program2([],[]).

convert_program2([L1|RL1s],[L7|RL2s]):-
	free_variables(L1,L2),
	new_program_vars(L2,L3,1,N),
	replace_program_vars(L1,L4),
	replace_program_vars(L3,L5),
	replace(L4,L5,L6,[],_),
	rectify_clause(L6,L7,N),
	convert_program2(RL1s,RL2s).

explicit_unification(Prog,Prog3):-
	vars(Prog,Vars),
        next_var_number(Vars,0,N),
        repeated_vars(Vars,Vars,Vars2),
        new_substitution(Vars2,N,Vars3),
        new_equality(Vars3,Eq),
        subst_vars(Vars3,_,Prog,Prog2),
        union(Prog2,Eq,Prog3).

subst_vars([(var_(X),var_(Y))|R],R1,[var_(X)|R2],[var_(Y)|R3]):-!,
	subst_vars(R,R1,R2,R3).

subst_vars([(var_(X),var_(Y))|R],R1,[var_(Z)|R2],[var_(Z)|R3]):-!,
	subst_vars([(var_(X),var_(Y))|R],R1,R2,R3).

subst_vars([V|RV],RV2,[X|R1],[Y|R2]):-
	X=..[A|C1],
	subst_vars([V|RV],RV1,C1,C2),
        Y=..[A|C2],!,	
	subst_vars(RV1,RV2,R1,R2).

subst_vars(R,R,L,L).


dele(_,[],[],[]).

dele(X,[(X,Y)|R1],R2,[(X,Y)|R3]):-!,
	dele(X,R1,R2,R3).

dele(X,[(Z,Y)|R1],[(Z,Y)|R2],R3):-
	dele(X,R1,R2,R3).

new_equality([],[]).

new_equality([(var_(X),var_(Y))|R],R5):-
	dele(var_(X),R,R1,R2),
        new_equality2(var_(Y),R2,R3),
        new_equality(R1,R4),
        union(R3,R4,R5).

new_equality2(_,[],[]).

new_equality2(var_(Y),[(_,var_(Z))|R],[var_(Y)=var_(Z)|R2]):-
	new_equality2(var_(Y),R,R2).


new_substitution([],_,[]).

new_substitution([var_(X)|R],N,[(var_(X),var_(N))|R2]):-
             N1 is N + 1,
             new_substitution(R,N1,R2).


next_var_number([],N,N1):-
	N1 is N +1.
next_var_number([var_(N)|R],N1,N2):-
             N>=N1,
             next_var_number(R,N,N2).
next_var_number([var_(N)|R],N1,N2):-
             N<N1,
             next_var_number(R,N1,N2).


repeated_vars([],_,[]).

repeated_vars([X|R1],V,[X|R2]):-
	number_in(X,V,N),
        N>1,!,
        repeated_vars(R1,V,R2).

repeated_vars([_|R1],V,R2):-
        repeated_vars(R1,V,R2).


number_in(_,[],0).
number_in(X,[X|R],N):-!,
        number_in(X,R,N1),
	N is N1 + 1.

number_in(X,[_|R],N):-
        number_in(X,R,N).




rectify_clause([P1|R],[P2|R2],N):-
	P1=..[F|Args],
	rectify_args(Args,NArgs,NGoal,N),
	P2=..[F|NArgs],
	union(R,NGoal,R2).


rectify_args([],[],[],_).

rectify_args([var_(X)|RA],[var_(X)|RB],RG,N):-!,
	rectify_args(RA,RB,RG,N).

rectify_args([X|RA],[var_(N)|RB],[var_(N)=X|RG],N):-
	N1 is N +1,
	rectify_args(RA,RB,RG,N1).


replace_program_vars([],[]):-!.

replace_program_vars(L1,L2):-
	term_to_atom(L1,A),
	string_to_list(A,A1),
	var_conversion(A1,A2),
	string_to_list(A3,A2),
	atom_to_term(A3,L2,_).


var_conversion([],[]).

var_conversion([95,71|R1],[103|R2]):-!,
	var_conversion(R1,R2).

var_conversion([X|R1],[X|R2]):-
	var_conversion(R1,R2).


new_program_vars([],[],N,N).

new_program_vars([A|RV1],[(A,var_(N))|RV2],N,N2):-
	N1 is N + 1,
	new_program_vars(RV1,RV2,N1,N2).




infer_types(File,Types4,DTypes,TP):-
	read_source_file(File,L1),
	convert_program(L1,L2),
	main_predicate(L2,MPred,NArgs),
	group_by_functors(L2,[],L3),
	group_program(L3,L4,Types1),
	flatten(Types1,Types2),
	built_in(BT),!,
	union(DTypes,Types2,Types3),
	include(TP,Types3,Types3b),
	union(Types3b,BT,Types3c),
	supertypes(L4,MPred,Types3c,Types4,1,_),
	warning_producer(MPred,NArgs,Types4).


warning_producer(MPred,NArgs,Types):-
	all_types_main(MPred,NArgs,MTypes),
	warning_producer2(MTypes,Types).

all_types_main(_,0,[]).

all_types_main(MPred,N,[type_(MPred,N)|R]):-
	N > 0,
	N1 is N - 1,
	all_types_main(MPred,N1,R).


warning_producer2([],_).

warning_producer2([type_(X,N)|RX],T):-
	member(type_(X,N)--->Y,T),
	warning_producer3(Y,T),!,
	nl,write('WARNING: Argument '),write(N),
	write(' of '),write(X),write(' deals with an empty document.'),
	warning_producer2(RX,T).

warning_producer2([_|RX],T):-
	warning_producer2(RX,T).


warning_producer3([],_).

warning_producer3([X|RX],T):-
	X=..[F,_,Con],
	F\=type_,
	empty_list(Con,T),
	warning_producer3(RX,T).

warning_producer3([X|RX],T):-
	X=..[type_,_,_],
	member(X--->Y,T),
	warning_producer3(Y,T),
	warning_producer3(RX,T).
	
empty_list([],_).

empty_list(type_(X,Y),T):-
	member(type_(X,Y)--->L,T),
	empty_list2(L,T).

empty_list2([],_).

empty_list2([[]],_).

empty_list2([X|RX],T):-
	member(X--->L,T),
	empty_list2(L,T),
	empty_list2(RX,T).


main_predicate([[X|_]|_],F,NArgs):-
	X=..[F|Args],
	length(Args,NArgs).


include([],L,L).

include([X--->Y|R],L,L2):-
	member(X--->Z,L),
	delete(L,X--->Z,LT),
	union([X--->Y],LT,L1),
	include(R,L1,L2).


initial_types([P|_],T):-
	P=..[F|Args],
	type_def(F,Args,T1,1),
	univ_initial_types(T1,T).

univ_initial_types([],[]).

univ_initial_types([X|RX],[X--->[any]|RY]):-
	univ_initial_types(RX,RY).


group_program([],[],[]).

group_program([(_,[X|R])|RP1],[[X|R]|RP2],[T|RT]):-
	initial_types(X,T),
	group_program(RP1,RP2,RT).


group_by_functors([],L,L).

group_by_functors([[H|R]|RC],L1,L3):-
	H=..[F|_],
	member((F,LF),L1),!,
	union(LF,[[H|R]],LF2),
	subst((F,LF),(F,LF2),L1,L2),
	group_by_functors(RC,L2,L3).

group_by_functors([[H|R]|RC],L1,L3):-
 	H=..[F|_],
	union(L1,[(F,[[H|R]])],L2),
	group_by_functors(RC,L2,L3).


reached_limit(T):-
	reached_limit1(T).

reached_limit(T):-
	reached_limit2(T).


reached_limit1(T):-
	member(_--->L,T),
	member(Y,L),
	Y=..[F|_],
	max_func(F,Y,0).

max_func(_,_,10):-!.

max_func(F,X,N):-
	X=..[F|[Args]],
	N1 is N + 1,
	max_func(F,Args,N1).


reached_limit2(Types):-
	member(X--->_,Types),
	reached_limit3(X,Types,[],0).

reached_limit3(_,_,_,10):-!.

reached_limit3(X,T,E,N):-
	member(X--->T1,T),
	type_names(T1,TypesT1),
	remov_types(TypesT1,E,TypesT2),
	union([X],E,E1),
	member(Y,TypesT2),
	N1 is N + 1,
	reached_limit3(Y,T,E1,N1).


remov_types(T,[],T).

remov_types([],_,[]).

remov_types([X|R1],E,R2):-
	member(X,E),!,
	remov_types(R1,E,R2).

remov_types([X|R1],E,[X|R2]):-
	remov_types(R1,E,R2).



type_names([],[]).

type_names([type_(X,Y)|R],[type_(X,Y)|R2]):-!,
	type_names(R,R2).

type_names([X|R],R3):-!,
	X=..[_|C],
	type_names(C,R1),	
	type_names(R,R2),
	union(R1,R2,R3).

