:-module(preprocess,[pre_process/3]).
:-op(800,xfx,--->).

%-------------------------------------------------------------------------
%
%   PRE PROCESSING
%
% by Jorge Coelho  jcoelho@dei.isep.ipp.pt
%    Mrio Florido amf@ncc.up.pt
%
%-------------------------------------------------------------------------

pre_process(File,Preds,Types):-
	infertypes:built_in_temp(L),
	assert(infertypes:built_in(L)),
	read_write_file(File,FN),
	new_dtd_types(FN,Preds,Types).


new_dtd_types([],[],[]):-!.

new_dtd_types([(Pred,DTD1,DTD2,N1,N2)|R1],Preds,Types1):-
	atom(DTD1),!,
	list_types(DTD1,T1),
	new_index(T1,T2,N1),
	atom(DTD2),
	list_types(DTD2,T3),
	new_index(T3,T4,N2),
	new_dtd_types(R1,P,T5),
	union([(Pred,N1,N2)],P,Preds),
	union(T2,T4,Types),
	union(Types,T5,Types1).

new_dtd_types(_,_,[]):-
	nl,write('ERROR: DTD filenames are not sufficiently instantiated'),nl.

new_index(T1,T3,N):-
	new_index2(T1,T2,N,S),
	new_index3(T2,T3,S).

new_index2([],[],_,[]).

new_index2([type_(X,Y)--->L|R1],[type_(X,N)--->L|R2],N,[(type_(X,Y),type_(X,N))|R3]):-
	new_index2(R1,R2,N,R3).

new_index3([],[],_).

new_index3([X--->Y1|R1],[X--->Y2|R2],S):-
	replace(Y1,S,Y2,[],_),
	new_index3(R1,R2,S).

read_write_file(File1,FN2):-
	open(File1,read,Fd),
	open('temp.xmlprolog',write,FdW),
	read(Fd,First),
	read_write_data(First,Fd,FdW,FN,1,_),
	flatten(FN,FN2),
	close(Fd),
	close(FdW).

read_write_data(end_of_file,_,_,[],N,N):-!.

read_write_data((_:-B),Fd,FdW,[FN|RFN],N1,N3):-
	tuple_to_list(B,C),
	verify_built_in(C,FN,N1,N2),
	read(Fd,Next),!,
	read_write_data(Next,Fd,FdW,RFN,N2,N3).

read_write_data((A:-B),Fd,FdW,RFN,N1,N2):-
	read(Fd,Next),!,
	write(FdW,A),write(FdW,':-'),write(FdW,'\n\t'),write(FdW,B),write(FdW,'.\n'),
	read_write_data(Next,Fd,FdW,RFN,N1,N2).

read_write_data((:-A),Fd,FdW,FN,N1,N2):-
	read(Fd,Next),!,
	write(FdW,':-'),write(FdW,A),write(FdW,'.\n'),
	read_write_data(Next,Fd,FdW,FN,N1,N2).

read_write_data((A),Fd,FdW,FN,N1,N2):-
	read(Fd,Next),
	write(FdW,A),write(FdW,'.\n'),
	read_write_data(Next,Fd,FdW,FN,N1,N2).

verify_built_in([],[],N,N).

verify_built_in(L,[(Process,DTD1,DTD2,N1,N2)],N1,N3):-
	N2 is N1 + 1,
	N3 is N2 + 1,
	member(X,L),
	member(Y,L),
	member(Z,L),
	X=..[xml_to_prolog,_,DTD1,_,_],
	Y=..[Process,_,_],
	Z=..[prolog_to_xml,DTD2,_,_,_].




