%:-module(xmlprolog,[xml_to_prolog/4,prolog_to_xml/4,main/2]).
:-op(800,xfx,--->).
:-use_module(xml2prolog).
:-use_module(prolog2xml).
:-use_module(infertypes).
:-use_module(dtdtypes).
:-use_module(replace).
:-use_module(preprocess).

%-------------------------------------------------------------------------
%
%   XML_TO_PROLOG & PROLOG_TO_XML
%
% by Jorge Coelho  jcoelho@dei.isep.ipp.pt
%    Mrio Florido amf@ncc.up.pt
%
%-------------------------------------------------------------------------

xml_to_prolog(XMLFile,DTDFile,XMLTerm1,DTDType):-
	xml2pro(XMLFile,DTDFile,XMLTerm1),
	list_types(DTDFile,DTDType).

prolog_to_xml(DTDFile,DTDType,XMLTerm,XMLFile):-
	list_types(DTDFile,DTDType),
	pro2xml(XMLTerm,XMLFile).


verify_inclusion([],_).

verify_inclusion([(P,N1,N2)|R],TR):-
	T1 =..[type_,P,1],
	TR1 =..[type_,root,N1],
	verify_subset(T1,TR1,TR),
	T2 =..[type_,P,2],
	TR2 =..[type_,root,N2],
	verify_subset(T2,TR2,TR),
	verify_inclusion(R,TR).


top_types([],[]).

top_types([(P,N1,N2)|R],[type_(P,N1)--->[type_(root,N1)],type_(P,N2)--->[type_(root,N2)]|R2]):-
	top_types(R,R2).

verify_subset(X,Y,T):-
	my_subset(X,Y,T),!.

verify_subset(type_(X,N),_,_):-
	nl,write('ERROR: Argument '),write(N),write(' of '),write(X),write(' incompatible with DTD type.'),nl,fail.

x-compile(File):-
	pre_process(File,Preds,BT),
        top_types(Preds,TP),
	infer_types('temp.xmlprolog',_,BT,TP),
        consult(File).


